/*
 * Frame.cpp
 *
 *  Created on: March, 2016
 *      Author: anonymous ECCV submission 1747
 */

#include "Frame.h"

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/highgui/highgui_c.h>
#include <opencv2/imgproc/imgproc.hpp>
#include <sys/types.h>
#include <cassert>
#include <sstream>
#include <vector>

#include "Skeleton.h"
#include "Video.h"

namespace nl_uu_science_gmt
{

void Frame::init(
		const int index, const SFramePtr &self)
{
	m_index = index;
	m_id = 0;
	m_self = self;

	m_timestamp = 0;
	m_image_filename = "";

	m_video = nullptr;
	m_next_frame = nullptr;
	m_this_frame = nullptr;
	m_prev_frame = nullptr;

	m_rotation = nullptr;
	m_translation = nullptr;

	bool m_flip = false;
	bool m_border = false;
}

void Frame::write(
		cv::FileStorage &fs) const
{
	assert(fs.isOpened());

	fs << "{";
	fs << "id" << (int) m_id;
	fs << "timestamp" << (int) m_timestamp;
	fs << "image_filename" << m_image_filename;
	if (m_rotation != nullptr && !m_rotation->empty())
		fs << "rotation" << *m_rotation;
	if (m_translation != nullptr && !m_translation->empty())
		fs << "translation" << *m_translation;
	fs << "skeletons_amount" << (int) m_skeletons.size();
	if (!m_skeletons.empty())
	{
		fs << "skeletons";
		fs << "{";
		for (auto skeleton : m_skeletons)
		{
			std::stringstream skeleton_name;
			skeleton_name << "skeleton_" << skeleton->getId();
			fs << skeleton_name.str() << *skeleton;
		}
		fs << "}";
	}

	if (!m_flow_filename.empty())
	{
		fs << "flow";
		fs << "{";
		fs << "filename" << m_flow_filename;
		fs << "x_size" << m_flow_size_x;
		fs << "x_type" << m_flow_type_x;
		fs << "y_size" << m_flow_size_y;
		fs << "y_type" << m_flow_type_y;
		fs << "}";
	}

	fs << "}";
}

void Frame::read(
		const cv::FileNode &fn)
{
	assert(fn.type() == cv::FileNode::MAP);

	int id = 0;
	fn["id"] >> id;
	m_id = id;
	int timestamp = 0;
	fn["timestamp"] >> timestamp;
	m_timestamp = timestamp;
	fn["image_filename"] >> m_image_filename;
	cv::Mat rotation, translation;
	fn["rotation"] >> rotation;
	m_rotation = std::make_shared<cv::Mat>(rotation);
	fn["translation"] >> translation;
	m_translation = std::make_shared<cv::Mat>(translation);

	uint index = 0;
	for (cv::FileNodeIterator n = fn["skeletons"].begin(); n != fn["skeletons"].end(); ++index)
	{
		auto skeleton = std::make_shared<Skeleton>();
		skeleton->init(index, skeleton);
		skeleton->setFrame(this);
		skeleton->setScreenWidth(m_video->getSize().width);
		skeleton->setFlipped(m_video->isMetaFlipped());
		n >> *skeleton;
		m_skeletons.push_back(skeleton);
	}

	if (!fn["flow"].empty())
	{
		for (cv::FileNodeIterator n = fn["flow"].begin(); n != fn["flow"].end();)
		{
			if ((*n).name() == "filename")
			{
				cv::String filename;
				n >> filename;
				m_flow_filename = filename;
			}
			else if ((*n).name() == "x_size")
				n >> m_flow_size_x;
			else if ((*n).name() == "x_type")
				n >> m_flow_type_x;
			else if ((*n).name() == "y_size")
				n >> m_flow_size_y;
			else if ((*n).name() == "y_type")
				n >> m_flow_type_y;
		}
	}
	else
	{
		m_flow_filename = "";
		m_flow_size_x = cv::Size(0, 0);
		m_flow_size_y = cv::Size(0, 0);
		m_flow_type_x = 0;
		m_flow_type_y = 0;
	}

	setup();
}

void Frame::setup()
{

}

void Frame::getImage(
		cv::Mat &image) const
{
	if (m_frame_image.empty())
	{
		if (m_video->getVideoCapture() != nullptr)
		{
			m_video->getVideoCapture()->set(CV_CAP_PROP_POS_FRAMES, (double) m_index);
			*m_video->getVideoCapture() >> image;
			assert(!image.empty());
		}
		else
		{
			std::stringstream image_file;
			image_file << m_video->getVideoPath() << m_video->getImagePath() << m_image_filename;
			const std::string filename = image_file.str();
			image = cv::imread(filename);
			assert(!image.empty());
		}

		if (m_video->isFlipped())
			cv::flip(image, image, 1);
	}
	else
	{
		image = m_frame_image;
	}
}

void Frame::getImageCopy(
		cv::Mat &image) const
{
	cv::Mat or_img;
	getImage(or_img);
	image = or_img.clone();
}

void Frame::releaseImage()
{
	m_frame_image.release();
}

void Frame::addSkeleton(
		const Skeleton &skeleton)
{
	m_skeletons.push_back(std::make_shared<Skeleton>(skeleton));
}

std::deque<SFramePtr> Frame::advance(
		int frame_amount, int count) const
{
	std::deque<SFramePtr> frames;
	if (count < frame_amount && m_next_frame != nullptr)
		frames = m_next_frame->advance(frame_amount, ++count);
	if (count > 1)
		frames.push_front(m_this_frame);
	return frames;
}

std::deque<SFramePtr> Frame::regress(
		int frame_amount, int count) const
{
	std::deque<SFramePtr> frames;
	if (count < frame_amount && m_prev_frame != nullptr)
		frames = m_prev_frame->regress(frame_amount, ++count);
	if (count > 1)
		frames.push_back(m_this_frame);
	return frames;
}

}
/* namespace nl_uu_science_gmt */
