/*
 * Skeleton.cpp
 *
 *  Created on: March, 2016
 *      Author: anonymous ECCV submission 1747
 */

#include "Skeleton.h"

#include <opencv2/core/core_c.h>
#include <algorithm>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <iterator>
#include <sstream>
#include <string>

#include "Frame.h"
#include "Video.h"

namespace nl_uu_science_gmt
{

const JointNameMap Skeleton::JointNames = initJointNames();
const Kinect2JointNameMap Skeleton::Kinect2JointNames = initKinect2JointNames();
const ActionNameMap Skeleton::ActionNames = initActionNames();
const Kinect12Map Skeleton::KinectSynonyms = initKinect12Map();

void Skeleton::init(const int index, const SSkelPtr &self)
{
	m_id = 0;
	m_index = index;
	m_self = self;
	m_fid = 0;
	m_action_class = AC_NA;
	m_flipped = false;
	m_screen_width = 0;
	m_angle = 0;

	m_rotation = nullptr;
	m_translation = nullptr;
	m_camera_matrix = nullptr;
	m_dist_coeff = nullptr;

	m_frame = nullptr;
}

void Skeleton::write(cv::FileStorage &fs) const
{
	assert(fs.isOpened());

	fs << "{";
	fs << "id" << (int) m_id;
	fs << "fid" << (int) m_fid;
	fs << "aid" << (int) m_action_class;
	fs << "action_name" << ActionNames.at(m_action_class);
	if(!m_points2D.empty()) fs << "joints_amount" << (int) m_points2D.size();
	if(!m_k2_points2D.empty()) fs << "joints_amount" << (int) m_k2_points2D.size();
	fs << "joints";
	fs << "{";
	for(auto joint2d : m_points2D)
	{
		std::stringstream joint_name;
		joint_name << "joint_" << joint2d.first;
		fs << joint_name.str();
		fs << "{";
		fs << "id" << (int) joint2d.first;
		fs << "name" << JointNames.at(joint2d.first);
		fs << "point2Dd" << joint2d.second;
		auto joint3d = m_points3D.find(joint2d.first);
		if(joint3d != m_points3D.end())
			fs << "point3Dd" << joint3d->second;
		fs << "}";
	}
	for(auto joint2d : m_k2_points2D)
	{
		std::stringstream joint_name;
		joint_name << "joint_" << joint2d.first;
		fs << joint_name.str();
		fs << "{";
		fs << "id" << (int) joint2d.first;
		fs << "name" << Kinect2JointNames.at(joint2d.first);
		fs << "point2Dd" << joint2d.second;
		auto joint3d = m_k2_points3D.find(joint2d.first);
		if(joint3d != m_k2_points3D.end())
			fs << "point3Dd" << joint3d->second;
		fs << "is_inferred" << (m_k2_inferred_joints.find(joint2d.first) != m_k2_inferred_joints.end());
		fs << "}";
	}
	fs << "}";
	fs << "}";
}

void Skeleton::read(const cv::FileNode &fn)
{
	assert(fn.type() == cv::FileNode::MAP);
	assert(m_screen_width > 0);

	int id = 0;
	fn["id"] >> id;
	m_id = id;
	int fid = 0;
	fn["fid"] >> fid;
	m_fid = fid;
	int aid = 0;
	fn["aid"] >> aid;
	if (aid > 0) m_action_class = (ACTION_CLASSES_INDEX) aid;

	double min_x = DBL_MAX;
	double min_y = DBL_MAX;
	double max_x = -DBL_MAX;
	double max_y = -DBL_MAX;

	for (cv::FileNodeIterator n = fn["joints"].begin(); n != fn["joints"].end(); ++n)
	{
		assert((*n).type() == cv::FileNode::MAP);

		int id = 0;
		(*n)["id"] >> id;

		if (m_frame != nullptr && m_frame->getVideo()->getVersion() == ShakeFiveVersion_Kinect1)
		{
			auto nui_id = (NUI_SKELETON_POSITION_INDEX) id;

			for (auto node = (*n).begin(); node != (*n).end(); ++node)
			{
				if ((*node).name() == "point2Di" || (*node).name() == "point2Dd")
				{
					cv::Point2d point2Dd;
					(*node) >> point2Dd;
					if (m_flipped) point2Dd.x = m_screen_width - point2Dd.x;
					m_points2D.insert(std::make_pair(nui_id, point2Dd));

					if (m_points2D.at(nui_id).x < min_x) min_x = m_points2D.at(nui_id).x;
					if (m_points2D.at(nui_id).x > max_x) max_x = m_points2D.at(nui_id).x;
					if (m_points2D.at(nui_id).y < min_y) min_y = m_points2D.at(nui_id).y;
					if (m_points2D.at(nui_id).y > max_y) max_y = m_points2D.at(nui_id).y;
				}
				else if ((*node).name() == "point3Df" || (*node).name() == "point3Dd")
				{
					cv::Point3d point3Dd;
					(*node) >> point3Dd;
					//FIXME Kinect3D is not flipped if the image was flipped...
					m_points3D.insert(std::make_pair(nui_id, point3Dd));
				}
			}
		}
		else
		{
			auto jt_id = (Kinect2JointType) id;

			cv::Point2d point2Dd;
			(*n)["point2Dd"] >> point2Dd;
			if (m_flipped) point2Dd.x = m_screen_width - point2Dd.x;
			m_k2_points2D.insert(std::make_pair(jt_id, point2Dd));

			if (m_k2_points2D.at(jt_id).x < min_x) min_x = m_k2_points2D.at(jt_id).x;
			if (m_k2_points2D.at(jt_id).x > max_x) max_x = m_k2_points2D.at(jt_id).x;
			if (m_k2_points2D.at(jt_id).y < min_y) min_y = m_k2_points2D.at(jt_id).y;
			if (m_k2_points2D.at(jt_id).y > max_y) max_y = m_k2_points2D.at(jt_id).y;

			cv::Point3d point3Dd;
			(*n)["point3Dd"] >> point3Dd;
			m_k2_points3D.insert(std::make_pair(jt_id, point3Dd));

			bool is_inferred;
			(*n)["is_inferred"] >> is_inferred;
			if (is_inferred) m_k2_inferred_joints.insert((Kinect2JointType) id);
		}
	}

	if (m_frame != nullptr)
	{
		assert(m_frame->getVideo() != nullptr);
		m_version = m_frame->getVideo()->getVersion();
	}
	if (m_version >= ShakeFiveVersion_Kinect2)
	{
		for (auto joint : m_k2_points2D)
		{
			auto qj = joint.first;
			auto k1joint_it = std::find_if(Skeleton::KinectSynonyms.begin(), Skeleton::KinectSynonyms.end(), [qj](const std::pair<NUI_SKELETON_POSITION_INDEX, Kinect2JointType> &joint){
					return qj == joint.second;
			});
			if (k1joint_it != Skeleton::KinectSynonyms.end())
			{
				m_points2D.insert(std::make_pair(k1joint_it->first, joint.second));
			}
		}

		for (auto joint : m_k2_points3D)
		{
			auto qj = joint.first;
			auto k1joint_it = std::find_if(Skeleton::KinectSynonyms.begin(), Skeleton::KinectSynonyms.end(), [qj](const std::pair<NUI_SKELETON_POSITION_INDEX, Kinect2JointType> &joint){
					return qj == joint.second;
			});
			if (k1joint_it != Skeleton::KinectSynonyms.end())
			{
				m_points3D.insert(std::make_pair(k1joint_it->first, joint.second));
			}
		}
	}
	assert(!m_points2D.empty());

	m_bbox = cv::Rect(min_x, min_y, max_x - min_x, max_y - min_y);

	calcDirection();
}

void Skeleton::drawBone(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const NUI_SKELETON_POSITION_INDEX &joint0, const NUI_SKELETON_POSITION_INDEX &joint1,
		const cv::Scalar &color, cv::Mat &frame)
{
	auto skeleton_joint0 = skeleton.find(joint0);
	if (skeleton_joint0 != skeleton.end())
	{
		auto skeleton_joint1 = skeleton.find(joint1);
		if (skeleton_joint1 != skeleton.end())
		{
			line(frame, skeleton_joint0->second, skeleton_joint1->second, color, 2, CV_AA);
		}
	}
}

void Skeleton::drawK2Bone(const std::map<Kinect2JointType, cv::Point2d> &skeleton, const Kinect2JointType &joint0,
		const Kinect2JointType &joint1, const cv::Scalar &color, cv::Mat &frame)
{
	auto skeleton_joint0 = skeleton.find(joint0);
	if (skeleton_joint0 != skeleton.end())
	{
		auto skeleton_joint1 = skeleton.find(joint1);
		if (skeleton_joint1 != skeleton.end())
		{
			if(m_k2_inferred_joints.find(skeleton_joint0->first) == m_k2_inferred_joints.end()
				&& m_k2_inferred_joints.find(skeleton_joint1->first) == m_k2_inferred_joints.end())
			{
				cv::Point2d start(m_flipped ?
						m_screen_width - skeleton_joint0->second.x : skeleton_joint0->second.x, skeleton_joint0->second.y);
				cv::Point2d end(m_flipped ?
						m_screen_width - skeleton_joint1->second.x : skeleton_joint1->second.x, skeleton_joint1->second.y);
				line(frame, start, end, color, 2, CV_AA);
			}
		}
	}
}

void Skeleton::drawTorso(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawBone(skeleton, SKELETON_POSITION_HEAD, SKELETON_POSITION_SHOULDER_CENTER, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SHOULDER_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SHOULDER_RIGHT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SPINE, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SPINE, SKELETON_POSITION_HIP_CENTER, color, frame);
	drawBone(skeleton, SKELETON_POSITION_HIP_CENTER, SKELETON_POSITION_HIP_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_HIP_CENTER, SKELETON_POSITION_HIP_RIGHT, color, frame);
}

void Skeleton::drawLeftArm(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_LEFT, SKELETON_POSITION_ELBOW_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_ELBOW_LEFT, SKELETON_POSITION_WRIST_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_WRIST_LEFT, SKELETON_POSITION_HAND_LEFT, color, frame);
}

void Skeleton::drawRightArm(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_RIGHT, SKELETON_POSITION_ELBOW_RIGHT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_ELBOW_RIGHT, SKELETON_POSITION_WRIST_RIGHT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_WRIST_RIGHT, SKELETON_POSITION_HAND_RIGHT, color, frame);
}

void Skeleton::drawLeftLeg(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawBone(skeleton, SKELETON_POSITION_HIP_LEFT, SKELETON_POSITION_KNEE_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_KNEE_LEFT, SKELETON_POSITION_ANKLE_LEFT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_ANKLE_LEFT, SKELETON_POSITION_FOOT_LEFT, color, frame);
}

void Skeleton::drawRightLeg(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawBone(skeleton, SKELETON_POSITION_HIP_RIGHT, SKELETON_POSITION_KNEE_RIGHT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_KNEE_RIGHT, SKELETON_POSITION_ANKLE_RIGHT, color, frame);
	drawBone(skeleton, SKELETON_POSITION_ANKLE_RIGHT, SKELETON_POSITION_FOOT_RIGHT, color, frame);
}

void Skeleton::drawK2Torso(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawK2Bone(skeleton, JointType_K2_Head, JointType_K2_Neck, color, frame);
	drawK2Bone(skeleton, JointType_K2_Neck, JointType_K2_SpineShoulder, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineShoulder, JointType_K2_SpineMid, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineMid, JointType_K2_SpineBase, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineShoulder, JointType_K2_ShoulderRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineShoulder, JointType_K2_ShoulderLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineBase, JointType_K2_HipRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_SpineBase, JointType_K2_HipLeft, color, frame);
}

void Skeleton::drawK2LeftArm(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawK2Bone(skeleton, JointType_K2_ShoulderLeft, JointType_K2_ElbowLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_ElbowLeft, JointType_K2_WristLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_WristLeft, JointType_K2_HandLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_HandLeft, JointType_K2_HandTipLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_WristLeft, JointType_K2_ThumbLeft, color, frame);
}

void Skeleton::drawK2RightArm(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawK2Bone(skeleton, JointType_K2_ShoulderRight, JointType_K2_ElbowRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_ElbowRight, JointType_K2_WristRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_WristRight, JointType_K2_HandRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_HandRight, JointType_K2_HandTipRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_WristRight, JointType_K2_ThumbRight, color, frame);
}

void Skeleton::drawK2LeftLeg(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawK2Bone(skeleton, JointType_K2_HipLeft, JointType_K2_KneeLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_KneeLeft, JointType_K2_AnkleLeft, color, frame);
	drawK2Bone(skeleton, JointType_K2_AnkleLeft, JointType_K2_FootLeft, color, frame);
}

void Skeleton::drawK2RightLeg(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &color, cv::Mat &frame)
{
	drawK2Bone(skeleton, JointType_K2_HipRight, JointType_K2_KneeRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_KneeRight, JointType_K2_AnkleRight, color, frame);
	drawK2Bone(skeleton, JointType_K2_AnkleRight, JointType_K2_FootRight, color, frame);
}

void Skeleton::drawSkeleton(const std::map<NUI_SKELETON_POSITION_INDEX, cv::Point2d> &skeleton,
		const cv::Scalar &bone_color, cv::Mat &frame)
{
	auto color = Color_GRAY;
	auto bone_color_l = Color_RED;
	auto bone_color_r = Color_GREEN;
	if (bone_color != Color_BLACK)
	{
		bone_color_l = bone_color;
		bone_color_r = bone_color;
	}

	drawBone(skeleton, SKELETON_POSITION_HEAD, SKELETON_POSITION_SHOULDER_CENTER, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SHOULDER_LEFT, bone_color_l, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SHOULDER_RIGHT, bone_color_r, frame);
	drawBone(skeleton, SKELETON_POSITION_SHOULDER_CENTER, SKELETON_POSITION_SPINE, color, frame);
	drawBone(skeleton, SKELETON_POSITION_SPINE, SKELETON_POSITION_HIP_CENTER, color, frame);
	drawBone(skeleton, SKELETON_POSITION_HIP_CENTER, SKELETON_POSITION_HIP_LEFT, bone_color_l, frame);
	drawBone(skeleton, SKELETON_POSITION_HIP_CENTER, SKELETON_POSITION_HIP_RIGHT, bone_color_r, frame);

	drawLeftArm(skeleton, bone_color_l, frame);
	drawLeftLeg(skeleton, bone_color_l, frame);
	drawRightArm(skeleton, bone_color_r, frame);
	drawRightLeg(skeleton, bone_color_r, frame);
}

void Skeleton::drawK2Skeleton(const std::map<Kinect2JointType, cv::Point2d> &skeleton,
		const cv::Scalar &bone_color, cv::Mat &frame)
{
	auto color = Color_GRAY;
	auto bone_color_l = Color_RED;
	auto bone_color_r = Color_GREEN;
	if (bone_color != Color_BLACK)
	{
		bone_color_l = bone_color;
		bone_color_r = bone_color;
	}

	drawK2Torso(skeleton, color, frame);
	drawK2LeftArm(skeleton, bone_color_l, frame);
	drawK2LeftLeg(skeleton, bone_color_l, frame);
	drawK2RightArm(skeleton, bone_color_r, frame);
	drawK2RightLeg(skeleton, bone_color_r, frame);
}

void Skeleton::drawSkeleton(const cv::Scalar &joint_color, const cv::Scalar &bone_color, cv::Mat &frame,
		ShakeFiveVersion version)
{
	std::stringstream skeleton_info;
	skeleton_info << "a:" << ActionNames.at(m_action_class) << "  id:" << m_id;

	if(version == ShakeFiveVersion_Kinect2)
	{
		auto head_point = m_k2_points2D.find(JointType_K2_Head);
		if(head_point != m_k2_points2D.end())
			Video::putStrokedText(frame, skeleton_info.str(), head_point->second, CV_FONT_HERSHEY_SIMPLEX, .8, Color_CYAN, 2);

		drawK2Skeleton(m_k2_points2D, bone_color, frame);
		for (auto joint : m_k2_points2D)
		{
			cv::Point2d dot(m_flipped ? m_screen_width - joint.second.x : joint.second.x, joint.second.y);
			circle(frame, dot, 2, joint_color, -1, CV_AA);
		}
	}
	else
	{
		auto head_point = m_points2D.find(SKELETON_POSITION_HEAD);
		if(head_point != m_points2D.end())
			Video::putStrokedText(frame, skeleton_info.str(), head_point->second, CV_FONT_HERSHEY_SIMPLEX, .8, Color_CYAN, 2);

		drawSkeleton(m_points2D, bone_color, frame);
		for (auto joint : m_points2D)
			circle(frame, joint.second, 2, joint_color, -1, CV_AA);
	}
}

double Skeleton::calcDirection()
{
	auto ls_shoulder = m_points3D.find(SKELETON_POSITION_SHOULDER_LEFT);
	auto rs_shoulder = m_points3D.find(SKELETON_POSITION_SHOULDER_RIGHT);
	auto ls_hip = m_points3D.find(SKELETON_POSITION_HIP_LEFT);
	auto rs_hip = m_points3D.find(SKELETON_POSITION_HIP_RIGHT);

	if (m_angle == -1 && !(ls_shoulder == m_points3D.end() || rs_shoulder == m_points3D.end() ||
			ls_hip == m_points3D.end() || rs_hip == m_points3D.end()))
	{
		double dx = ls_shoulder->second.x - rs_shoulder->second.x;
		double dz = ls_shoulder->second.z - rs_shoulder->second.z;

		cv::Point3d chest_center;
		chest_center.x = (ls_shoulder->second.x + rs_shoulder->second.x) / 2;
		chest_center.y = (ls_shoulder->second.y + rs_shoulder->second.y) / 2;
		chest_center.z = (ls_shoulder->second.z + rs_shoulder->second.z) / 2;
		cv::Point3d chest_perpendicular;
		chest_perpendicular.x = chest_center.x - (dz / 2);
		chest_perpendicular.y = (ls_shoulder->second.y + rs_shoulder->second.y) / 2;
		chest_perpendicular.z = chest_center.z + (dx / 2);
		double chest_length = cv::norm(chest_perpendicular - chest_center);
		double sq_length_ab = chest_length * chest_length;
		cv::Point3d chest_normal;
		chest_normal.x = chest_center.x - chest_length;
		chest_normal.y = (ls_shoulder->second.y + rs_shoulder->second.y) / 2;
		chest_normal.z = (ls_shoulder->second.z + rs_shoulder->second.z) / 2;
		double norm_chest = cv::norm(chest_normal - chest_perpendicular);
		double sq_length_c = norm_chest * norm_chest;
		double chest_angle = acos(((sq_length_ab + sq_length_ab) - sq_length_c) / (2.0 * chest_length * chest_length));

		dx = ls_hip->second.x - rs_hip->second.x;
		dz = ls_hip->second.z - rs_hip->second.z;

		cv::Point3d hip_center;
		hip_center.x = (ls_hip->second.x + rs_hip->second.x) / 2;
		hip_center.y = (ls_hip->second.y + rs_hip->second.y) / 2;
		hip_center.z = (ls_hip->second.z + rs_hip->second.z) / 2;
		cv::Point3d hip_perpendicular;
		hip_perpendicular.x = hip_center.x - (dz / 2);
		hip_perpendicular.y = (ls_hip->second.y + rs_hip->second.y) / 2;
		hip_perpendicular.z = hip_center.z + (dx / 2);
		double hip_length = cv::norm(hip_perpendicular - hip_center);
		sq_length_ab = hip_length * hip_length;
		cv::Point3d hip_normal;
		hip_normal.x = hip_center.x - hip_length;
		hip_normal.y = (ls_hip->second.y + rs_hip->second.y) / 2;
		hip_normal.z = (ls_hip->second.z + rs_hip->second.z) / 2;
		double norm_hip = cv::norm(hip_normal - hip_perpendicular);
		sq_length_c = norm_hip * norm_hip;
		double hip_angle = acos(((sq_length_ab + sq_length_ab) - sq_length_c) / (2.0 * hip_length * hip_length));

		m_angle = (chest_angle + hip_angle) / 2;
	}

	return m_angle;
}

} /* namespace nl_uu_science_gmt */
