/*
 * Video.cpp
 *
 *  Created on: March, 2016
 *      Author: anonymous ECCV submission 1747
 */

#include "Video.h"

#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/core/core_c.h>
#include <opencv2/core/mat.hpp>
#include <opencv2/core/types_c.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/highgui/highgui_c.h>
#include <opencv2/imgproc/imgproc.hpp>
#include <sys/types.h>
#include <algorithm>
#include <cassert>
#include <cfloat>
#include <cstdio>
#include <iostream>
#include <iterator>
#include <sstream>
#include <utility>

#include "Frame.h"
#include "Skeleton.h"

namespace nl_uu_science_gmt
{

void Video::init(
		const int id)
{
	m_id = id;
	m_version = ShakeFiveVersion_Kinect2;
	m_video_path = "";
	m_video_metadata_filename = "";
	m_video_filename = "";
	m_image_path = "";
	m_stride = 0;
	m_video_data_flipped = false;

	m_video_capture = nullptr;
	m_video_writer = nullptr;
	m_camera_matrix = nullptr;
	m_dist_coeff = nullptr;

	m_paused = false;
	m_stop = false;
	m_next = false;
	m_current_fid = 0;
}

void Video::write(
		cv::FileStorage &fs) const
{
	assert(fs.isOpened());

	fs << "{";
	fs << "id" << (int) m_id;
	fs << "version" << (int) m_version;
	fs << "image_path" << m_image_path;
	fs << "video_size" << m_size;
	fs << "video_stride" << (int) m_stride;
	if (m_camera_matrix != nullptr)
		fs << "camera_matrix" << *m_camera_matrix;
	if (m_dist_coeff != nullptr)
		fs << "dist_coeff" << *m_dist_coeff;
	fs << "frames_amount" << (int) m_frames.size();
	fs << "is_flipped" << m_video_data_flipped;
	if (!m_frames.empty())
	{
		fs << "frames";
		fs << "{";
		for (auto frame : m_frames)
		{
			std::stringstream frame_name;
			frame_name << "frame_" << frame->getId();
			fs << frame_name.str() << *frame;
		}
		fs << "}";
	}
	fs << "}";
}

void Video::read(
		const cv::FileNode &fn)
{
	assert(fn.type() == cv::FileNode::MAP);

	m_frames.clear();

	int id = 0;
	fn["id"] >> id;
	m_id = id;

	int version = 0;
	fn["version"] >> version;
	if (version == 0)
		m_version = ShakeFiveVersion_Kinect1;
	else
		m_version = (ShakeFiveVersion) version;

	std::string image_path;
	if ((int) m_version > 2)
	{
		for (cv::FileNodeIterator n = fn["video_data"].begin(); n != fn["video_data"].end(); ++n)
		{
			if ((*n).name() == "file")
			{
				std::string video_file = (std::string) (*n);
				m_video_capture = std::make_shared<cv::VideoCapture>(m_video_path + video_file);
				assert(m_video_capture->isOpened());
				m_fps = m_video_capture->get(CV_CAP_PROP_FPS);
			}
			else if ((*n).name() == "type")
			{
				m_video_type = (std::string) (*n);
			}
			else if ((*n).name() == "codec")
			{
				m_video_codec = (std::string) (*n);
			}
			else if ((*n).name() == "pix_fmt")
			{
				m_video_pix_fmt = (std::string) (*n);
			}
			else if ((*n).name() == "is_flipped")
			{
				m_video_data_flipped = (int) (*n);
			}
		}

		for (cv::FileNodeIterator n = fn["image_data"].begin(); n != fn["image_data"].end(); ++n)
		{
			if ((*n).name() == "path")
			{
				image_path = (std::string) (*n);
			}
			else if ((*n).name() == "type")
			{
				m_image_type = (std::string) (*n);
			}
			else if ((*n).name() == "is_flipped")
			{
				m_image_data_flipped = (int) (*n);
			}
		}
	}
	else
	{
		if (fn["image_path"].empty())
		{
			for (cv::FileNodeIterator n = fn["image_data"].begin(); n != fn["image_data"].end(); ++n)
			{
				if ((*n).name() == "path")
				{
					image_path = (std::string) (*n);
				}
				else if ((*n).name() == "type")
				{
					m_image_type = (std::string) (*n);
				}
				else if ((*n).name() == "is_flipped")
				{
					m_image_data_flipped = (int) (*n);
				}
			}
		}
		else
		{
			assert(!fn["image_path"].empty());
			fn["image_path"] >> image_path;
			m_image_data_flipped = true;
		}
	}

	assert(!image_path.empty());
	if (image_path.substr(image_path.length() - 1, 1) == "\\")
		image_path.replace(image_path.length() - 1, 1, "/");
	else if (image_path.substr(image_path.length() - 1, 1) != "/")
		image_path += "/";
	m_image_path = image_path;

	fn["video_size"] >> m_size;
	int stride = 0;
	fn["video_stride"] >> stride;
	m_stride = stride;
	fn["is_flipped"] >> m_meta_flipped;
	int frames_amount;
	fn["frames_amount"] >> frames_amount;

	if (frames_amount > 0)
	{
		cv::Mat camera_matrix;
		fn["camera_matrix"] >> camera_matrix;
		cv::Mat dist_coeff;
		fn["dist_coeff"] >> dist_coeff;

		double min_x = DBL_MAX, min_y = DBL_MAX, min_z = DBL_MAX;
		double max_x = -DBL_MAX, max_y = -DBL_MAX, max_z = -DBL_MAX;

		std::vector<double> frame_timings;

		uint index = 0;
		SFramePtr prev_frame = nullptr;
		for (cv::FileNodeIterator n = fn["frames"].begin(); n != fn["frames"].end(); ++index)
		{
			auto frame = std::make_shared<Frame>();
			frame->init(index, frame);
			frame->setVideo(this);

			n >> *frame;

			if (frame->getTlb().x < min_x)
				min_x = frame->getTlb().x;
			if (frame->getBrf().x > max_x)
				max_x = frame->getBrf().x;
			if (frame->getTlb().y < min_y)
				min_y = frame->getTlb().y;
			if (frame->getBrf().y > max_y)
				max_y = frame->getBrf().y;
			if (frame->getTlb().z < min_z)
				min_z = frame->getTlb().z;
			if (frame->getBrf().z > max_z)
				max_z = frame->getBrf().z;

			initSkelColors(frame);

			frame->setVideo(this);
			frame->setPrevFrame(prev_frame);
			if (prev_frame != nullptr)
			{
				frame_timings.push_back(frame->getTimeStamp() - prev_frame->getTimeStamp());
				prev_frame->setNextFrame(frame);
			}

			m_frames.emplace_back(frame);
			prev_frame = frame;
		}

		const auto sum_ft = std::accumulate(frame_timings.begin(), frame_timings.end(), 0.);
		const auto mean_ft = sum_ft / (double) frame_timings.size();
		m_fps = 1 / (mean_ft * 1.0e-7);
		m_tlb = cv::Point3d(min_x, min_y, min_z);
		m_brf = cv::Point3d(max_x, max_y, max_z);
	}

	setup();
}

void Video::setup()
{
}

void Video::keyboard(
		const int key)
{
	switch (key)
	{
	case 27:
		m_stop = !m_stop;
		break;

	case 'q':
		m_next = !m_next;
		break;

	case 'p':
		m_paused = !m_paused;
		break;

	case 'b':
		--m_current_fid;
		break;

	case 'n':
		++m_current_fid;
		break;
	}
}

void Video::createVideoCapture(
		const std::string &filename)
{
	m_video_capture = std::make_shared<cv::VideoCapture>(filename);
}

void Video::initSkelColors(
		const SFramePtr &frame)
{
	for (auto skel : frame->getSkeletons())
	{
		auto sk_id = skel->getId();
		auto sk = m_skel_colors.find(sk_id);
		if (sk == m_skel_colors.end())
		{
			cv::Scalar color;
			switch (m_skel_colors.size())
			{
			case 1:
				color = Color_GREEN;
				break;
			case 2:
				color = Color_BLUE;
				break;
			case 3:
				color = Color_KHAKI;
				break;
			case 4:
				color = Color_PURPLE;
				break;
			case 5:
				color = Color_ORANGE;
				break;
			case 6:
				color = Color_BROWN;
				break;
			default:
				color = Color_RED;
				break;
			}
			m_skel_colors.insert(std::make_pair(sk_id, color));
		}
	}
}

void Video::createVideoWriter(
		const std::string &filename, const int fps)
{
	const int fourcc = CV_FOURCC('X', 'V', 'I', 'D');
	m_video_writer = std::make_shared<cv::VideoWriter>(filename, fourcc, fps, m_size, true);
}

void Video::addFrame(
		SFramePtr &frame)
{
	m_frames.push_back(frame);
}

void Video::writeFrame(
		const cv::Mat &image)
{
	cv::Mat w_image;
	if (image.size().area() != m_size.area())
	{
		cv::resize(image, w_image, m_size);
	}
	else
	{
		w_image = image;
	}

	*m_video_writer << w_image;
}

int Video::play()
{
	const std::string window_name = "video";
	const std::string trackbar_name = "frame";
	cv::namedWindow(window_name, CV_WINDOW_KEEPRATIO);
	cv::resizeWindow(window_name, m_size.width, m_size.height);
	cv::createTrackbar(trackbar_name, window_name, &m_current_fid, (int) m_frames.size() - 1);

	const int digits = 1;
	std::string st_sec = "0";

	int64 t = cv::getTickCount(), ot;
	char sece_str[50], rfps_str[50], mfps_str[50], skel_str[250];

	SFramePtr frame = m_frames[0];
	const auto s_t = frame->getTimeStamp();
	auto ov = s_t;
	while (!(m_stop || m_next))
	{
		cv::Mat canvas;
		frame->getImage(canvas);

		for (auto skeleton : frame->getSkeletons())
		{
			skeleton->setFlipped(m_video_data_flipped);
			skeleton->setScreenWidth(m_size.width);

			skeleton->drawSkeleton(Color_YELLOW, Color_BLACK, canvas, m_version);
		}

		double seconds = (frame->getTimeStamp() - s_t) * 1.0e-7;
		sprintf(sece_str, "%.2f sec", seconds);
		putStrokedText(canvas, sece_str, cv::Point(8, 24 + 32 * 2), CV_FONT_HERSHEY_PLAIN, 2, Color_WHITE, 2, CV_AA);

		cv::setTrackbarPos(trackbar_name, window_name, m_current_fid);

		ot = t;
		t = cv::getTickCount();
		const double dt = cv::getTickFrequency() / double(t - ot);
		const double dv = 1. / ((frame->getTimeStamp() - ov) * 1.0e-7);

		sprintf(mfps_str, "%.2f video fps", dv);
		sprintf(rfps_str, "%.2f real fps", dt);
		putStrokedText(canvas, mfps_str, cv::Point(8, 24 + 32 * 1), CV_FONT_HERSHEY_PLAIN, 2, Color_GREEN, 2, CV_AA);
		putStrokedText(canvas, rfps_str, cv::Point(8, 24 + 32 * 3), CV_FONT_HERSHEY_PLAIN, 2, Color_GREEN, 2, CV_AA);

		cv::imshow("video", canvas);
		keyboard(cv::waitKey(30));

		if (!m_paused)
			m_current_fid += 1;
		if (m_current_fid >= (int) m_frames.size())
		{
			m_paused = true;
			m_current_fid = 0;
		}
		if (m_current_fid < 0)
			m_current_fid = (int) m_frames.size() - 1;

		ov = frame->getTimeStamp();
		frame = m_frames[m_current_fid];
	}

	return (m_stop || m_next) ? m_stop ? 27 : m_next ? 'n' : -1 : -1;
}

} /* namespace nl_uu_science_gmt */
